package edu.swe622.fers.model;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.NumberFormat;

/**
 * Class for handling and formatting latitude and longitude data.
 * Only contains two variables, which are assumed to be in decimal degrees. 
 * Also contains a number of other static variables for formatting
 * other lat long data. 
 * @author Doug Gillespie
 *
 */
public class LatLong implements Serializable, Cloneable {

	static final long serialVersionUID = 1;
	
	public transient static final double MetersPerMile = 1852.;
	/**
	 *  latitude in decimal degrees (North is positive; -90 <= latitude <= +90)
	 */
	protected double latitude;
	/**
	 *  longitude in decimal degrees (East is positive; -180 < longitude <= +180)
	 */
	protected double longitude;
	
	
	static public final int FORMAT_DECIMALMINUTES = 1;
	static public final int FORMAT_MINUTESSECONDS = 2;
	static private int formatStyle = FORMAT_DECIMALMINUTES;
	
	
	/**
	 * @return Returns the formatStyle.
	 */
	public static int getFormatStyle() {
		return formatStyle;
	}

	/**
	 * @param formatStyle The formatStyle to set.
	 */
	public static void setFormatStyle(int formatStyle) {
		LatLong.formatStyle = formatStyle;
	}

	public LatLong(double latitude, double longitude) {
		this.latitude = latitude;
		this.longitude = longitude;
	}
	
	public LatLong() {
		this.latitude = 0;
		this.longitude = 0;
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public LatLong clone() {
		try {
			return (LatLong) super.clone();
		}
		catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return null;
	}
	/**
	 * @return Returns the latitude.
	 */
	public double getLatitude() {
		return latitude;
	}
	/**
	 * @param latitude The latitude to set.
	 */
	public void setLatitude(double latitude) {
		this.latitude = latitude;
	}
	/**
	 * @return Returns the longitude.
	 */
	public double getLongitude() {
		return longitude;
	}
	/**
	 * @param longitude The longitude to set.
	 */
	public void setLongitude(double longitude) {
		while (longitude > 180) longitude -= 360.;
		while (longitude <= -180) longitude += 360.; 
		this.longitude = longitude;
	}
	
	/**
	 * formats the latitude string
	 * @return formated latitude string
	 */
	public String formatLatitude()
	{
		return formatLatitude(latitude);
	}
	/**
	 * convert a double decimal degrees into a formatted latlong string
	 * @param latlong
	 * @return a string representation of a Latitude
	 */
	public static String formatLatitude(double latlong) {
		if (latlong < 0) {
			return formatLatLong(latlong) + " S";
		}
		else  {
			return formatLatLong(latlong) + " N";
		}		
	}
	
	/**
	 * formats the longitude as a string
	 * @return formatted longitude string
	 */
	public String formatLongitude()
	{
		return formatLongitude(longitude);
	}
	/**
	 * convert a double decimal degrees into a formatted latlong string
	 * @param latlong
	 * @return a string representation of a Longitude
	 */
	public static String formatLongitude(double latlong) {
		if (latlong < 0) {
			return formatLatLong(latlong) + " W";
		}
		else  {
			return formatLatLong(latlong) + " E";
		}
	}
	
	private static String formatLatLong(double latLong, int style) {
		switch (style) {
		case FORMAT_DECIMALMINUTES:
			return formatLatLongDDM(latLong);
		case FORMAT_MINUTESSECONDS:
			return formatLatLongDMS(latLong);
		}
		return null;
	}
	
	private static String formatLatLongDMS(double latLong) {
		int d = Math.abs(getSignedDegrees(latLong));
		int m = getIntegerMinutes(latLong);
		double s = getSeconds(latLong);
		return String.format("%2d%02d\'%04.1f\'\'", d, m, s);
	}
	
	private static String formatLatLongDDM(double latlong) {
		latlong = Math.abs(latlong);
		int degrees = (int) Math.floor(latlong);
		double minutes = (latlong - degrees) * 60;
		return String.format("%2d%06.3f\'", degrees, minutes);
	}
	
	private static String formatLatLong(double latlong) {
		return formatLatLong(latlong, formatStyle);
	}
	
	public static String doubleDegrees2StringDegs(double latlong){
		double degrees =  Math.signum(latlong)*Math.floor(Math.abs(latlong));
		NumberFormat formatter = new DecimalFormat("0");
		return formatter.format(degrees);
	}
	
	public static String doubleDegrees2StringMins(double latlong){
		double degrees =  Math.signum(latlong)*Math.floor(Math.abs(latlong));
		double minutes = Math.abs((latlong - degrees) * 60);
		NumberFormat formatter = new DecimalFormat(".0");
		return formatter.format(minutes);
	}
	
	public static int getSignedDegrees(double latLong) {
		return (int) (Math.signum(latLong)*Math.floor(Math.abs(latLong)));
	}
	
	public static double getDecimalMinutes(double latLong) {
		double degrees = getSignedDegrees(latLong);
		return Math.abs((latLong - degrees) * 60);
	}
	
	/*
	 * Get some slightly weird rounding errors in al this which tends
	 * to make minutes 0 and seconds = 60, when we really 
	 * want 1 minute and 0 seconds. So need to slightly bump up the
	 * minutes before flooring them, then in the seconds calculation
	 * assume that they can never be negative, so set a minumium
	 * value for them
	 */
	public static int getIntegerMinutes(double latLong) {
		return (int) Math.floor(getDecimalMinutes(latLong)+.00000001);
	}
	
	public static double getSeconds(double latLong) {
		double decMinutes = getDecimalMinutes(latLong);
		double intMinutes = getIntegerMinutes(latLong);
		return Math.max((decMinutes - intMinutes) * 60., 0.);
	}

	/**
	 * @param trueHeading
	 *            Heading relative to True North
	 * @param distance
	 *            Distance to travel in meters.
	 * @return New Latlong in decimal degrees
	 */
	public LatLong travelDistanceMeters(double trueHeading,
			double distance) {
		return TravelDistanceMiles(trueHeading, distance / MetersPerMile);
	}

	/**
	 * @param trueHeading
	 *            Heading relative to True North
	 * @param distance
	 *            Distance to travel in nautical miles.
	 * @return New Latlong in decimal degrees
	 */
	public LatLong TravelDistanceMiles(double trueHeading,
			double distance) {
		LatLong newLatLong = new LatLong(this.latitude,
				this.longitude);
		double angle = Math.PI / 180. * (90. - trueHeading); // easier to
																// think like
																// this !
		double longDistortion = Math.cos(this.latitude * Math.PI / 180.);
		newLatLong.longitude += Math.cos(angle) * distance / 60.
				/ longDistortion;
		newLatLong.latitude += Math.sin(angle) * distance / 60.;

		// System.out.println("LatLongDegrees, latDegs: " + newLatLong.latDegs +
		// " longDegs: " + newLatLong.longDegs);

		return newLatLong;
	}
	
	double metersToDegreeLatitude(double meters) {
		return meters / 60. / LatLong.MetersPerMile;
	}
	
	double metersToDegreeLongitude(double meters, double latitudeDegrees) {
		double fac = Math.cos(latitudeDegrees * Math.PI / 180.);
		return meters / 60. / LatLong.MetersPerMile / fac;
	}
	
	public LatLong addDistanceMeters(double addX, double addY) {
		LatLong newLatLong = this.clone();
		newLatLong.latitude += metersToDegreeLatitude(addY);
		newLatLong.longitude += metersToDegreeLongitude(addX, (this.latitude + newLatLong.latitude)/2.);
		return newLatLong;
	}
	/**
	 * Calculate the distance from this positions to some other
	 * position in miles.
	 * @param destination destiantion position
	 * @return distance in miles
	 */
	public double distanceToMiles(LatLong destination) {
		double dist = Math.pow(60 * (latitude - destination.latitude), 2);
		double meanLat = (latitude + destination.latitude) / 2 * Math.PI / 180;
		double dist2 = Math.pow(60 * Math.cos(meanLat) * (longitude - destination.longitude), 2);
		return Math.sqrt(dist + dist2);
	}
	/**
	 * Calculate the distance from this positions to some other
	 * position in metres.
	 * @param destination destiantion position
	 * @return distance in metres
	 */
	public double distanceToMetres(LatLong destination) {
		return distanceToMiles(destination) * 1852.0;
	}
	public double distanceToMetresX(LatLong destination) {
		double longDist = (destination.getLongitude() - getLongitude()) * 
			getLongDistanceCorr(destination.getLatitude(), getLatitude());
		return longDist * 1852.0 * 60;
	}
	public double distanceToMetresY(LatLong destination) {
		double latDist = destination.getLatitude() - getLatitude();
		return latDist * 1852.0 * 60;
	}

	public boolean equals(LatLong ll) {
		return (ll.latitude == latitude && ll.longitude == longitude);
	}
	
	/*
	 * REturns angle between two points in degreed (relative to North)
	 */
	public double bearingTo(LatLong destination) {
		double latDist = destination.getLatitude() - getLatitude();
		double longDist = (destination.getLongitude() - getLongitude()) * 
		getLongDistanceCorr(destination.getLatitude(), getLatitude());
		double angle = Math.atan2(latDist, longDist) * 180. / Math.PI;
		angle = 90-angle;
		if (angle < 0) angle += 360;
		return angle;
	}
	
	private double getLongDistanceCorr(double latitude) {
		return Math.cos(latitude * Math.PI / 180.);
	}
	private double getLongDistanceCorr(double latitude1, double latitude2) {
		return getLongDistanceCorr((latitude1+latitude2)/2);
	}
	
	/**
	 * Clculate the perpendicular distance from this LatLong
	 * to a line joining latLong1 and latLong2
	 * @return cross track error in metres
	 */
	public double getCrossTrackError(LatLong latLong1, LatLong latLong2) {
		double lineBearing = latLong1.bearingTo(latLong2);
		double hereBearing = latLong1.bearingTo(this);
		double angDiff = lineBearing - hereBearing;
		double distHere = latLong1.distanceToMetres(this);
		return Math.abs(distHere * Math.sin(angDiff * Math.PI / 180));
	}
	
}
